unit Main;

interface

{$DEFINE SHOW_FPS}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DXDraws, DXClass, DIB, LDGFX;

const
  SCREEN_W = 400;
  SCREEN_H = 300;

const
  TRANSPARENT = $FF00FF;

const
  BND_UPPER  =  $01;
  BND_LEFT   =  $02;
  BND_LOWER  =  $04;
  BND_RIGHT  =  $08;

const
  SHADE_IMAGES = ' SKY1 RAIN ';

const
  SHADE_COUNT = 3;
  SHADE_SUB = $18;

const
  X_SCROLL_SIDE =  125;
  Y_SCROLL_SIDE =  100;

const
  FP = 256;

const
  MARGIN = 8;


type
  TPlayerState = (psIdle, psWalk, psJump, psFall);

type
  TPlayer =
    record
      HDirection,
      VDirection: Integer;
      FPXPos,
      FPYPos: Integer;
      FPXVel,
      FPYVel: Integer;
      FPGravity: Integer;
      State: TPlayerState;
      Width,
      Height: Integer;
      AnimSpeed,
      AnimPos: Integer;
      LightX,
      LightY: Integer;
      LightAnimPos: Integer;
      XShift,
      YShift: Integer;
      BmpWidth,
      BmpHeight: Integer;
      Spacebar: Boolean;
    end;

type
  DropRec =
    record
      FPXPos, FPYPos: Integer;
      FPXVel, FPYVel: Integer;
      Shade: Integer;
    end;

type
  CandleRec =
    record
      XPos, YPos: Integer;
      Light: Integer;
      FlareCounter: Integer;
    end;

var
  Keys: array[0..255] of Boolean;

type
  TLayerPtr = ^TLayer;

type
  TGameState = (gsInit, gsLevel);

type
  TGameWindow = class(TDXForm)
    DXDraw: TDXDraw;
    DXImageList: TDXImageList;
    DXTimer: TDXTimer;
    procedure DXTimerTimer(Sender: TObject; LagCount: Integer);
    procedure FormActivate(Sender: TObject);
    procedure FormDeactivate(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormCreate(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormDestroy(Sender: TObject);
  private
    BackgroundColor: Integer;
    CurGameState: TGameState;
    CurLevel: Integer;
    SkyLayer: TLayer;
    BackLayer: TLayer;
    FrontLayer: TLayer;
    XView: Integer;
    YView: Integer;
    Player: TPlayer;
    Drops: array of DropRec;
    FPWind: Integer;
    LightPerc: Integer;
    Candles: array of CandleRec;
    procedure CreateShades;
    procedure NewGameState (gs: TGameState);
    function ReadMap (X, Y: Integer; Side: Boolean): Integer;
    procedure InitMap;
    procedure RunFrame;
    procedure DrawFrame;
    procedure DrawLayer (L: TLayerPtr; X, Y, W, H: Integer; XScroll, YScroll: Integer;
                         Shade: Integer = SHADE_COUNT - 1);
    procedure InitPlayer (MapX, MapY: Integer);
    procedure DrawPlayer;
    procedure RunPlayer;
    procedure CheckLevelBounds;
    procedure MoveScreen;
    procedure RunDrops;
    procedure DrawDrops;
    function CheckCol (x, y: Integer): Boolean;
  public
  end;

var
  GameWindow: TGameWindow;

implementation

{$R *.DFM}

function UpcaseStr (s: string): string;
  var
    i: Integer;
begin
  for i := 1 to Length (s) do
    s[i] := Upcase (s[i]);
  Result := s;
end;

function MakeRect (X, Y, W, H: Integer): TRect;
begin
  Result := Rect (X, Y, X + W - 1, Y + H - 1);
end;

function udiv (N, D: Integer): Integer;
begin
  if (N < 0) then
    Result := N div D - Byte (N mod D <> 0)
  else
    Result := N div D;
end;

procedure SwapInt (var x, y: Integer);
  var
    z: Integer;
begin
  z := x;
  x := y;
  y := z;
end;

procedure TGameWindow.DXTimerTimer(Sender: TObject; LagCount: Integer);
  var
    fps: Integer;
begin
  if not DXDraw.CanDraw then
    Exit;

  DXDraw.Surface.Fill (BackgroundColor);

  RunFrame;
  DrawFrame;

  with DXDraw.Surface.Canvas do
  begin
{$IFDEF SHOW_FPS}
    Font.Color := clWhite;
    Brush.Style := bsClear;
    fps := DXTimer.FrameRate;
    if fps > 0 then
      TextOut (8, 8, IntToStr (fps) + ' fps');
{$ENDIF}
    Release;
  end;

  DXDraw.Flip;

end;

procedure TGameWindow.CreateShades;
  var
    pci: TPictureCollectionItem;
    i, x, y, W, H, N: Integer;
    tsn: array of string;
    tsc: Integer;
    dib: TDIB;

  function ShadePixel (RGB, N, Max: Integer): Integer;
    var
      r, g, b: Integer;
  begin
    if RGB = TRANSPARENT then
      Result := RGB
    else
    begin
      r := (RGB shr 16) and $FF;
      g := (RGB shr 8) and $FF;
      b := RGB and $FF;
      r := (r + r * N div Max) div 2;
      g := (g + g * N div Max) div 2;
      b := (b + b * N div Max) div 2;
      r := r - SHADE_SUB * N;
      g := g - SHADE_SUB * N;
      b := b - SHADE_SUB * N;
      if r < 0 then r := 0;
      if g < 0 then g := 0;
      if b < 0 then b := 0;
      Result := (RGB and $FF000000) + (r shl 16) + (g shl 8) + b;
    end;
  end;

begin
  with DXImageList do
  begin
    tsc := 0;
    for i := 0 to Items.Count - 1 do
      if Pos (UpcaseStr (Items[i].Name), SHADE_IMAGES) > 0 then
      begin
        Items[i].TransparentColor := TRANSPARENT;
        SetLength (tsn, tsc + 1);
        tsn[tsc] := Items[i].Name;
        Inc (tsc);
      end;

    for i := 0 to tsc - 1 do
    begin
      for N := 0 to SHADE_COUNT - 1 do
      begin
        pci := DXImageList.Items.Find (tsn[i]);
        W := pci.PatternWidth;
        H := pci.PatternHeight;
        dib := TDIB.Create;
        dib.Assign (pci.Picture.Graphic);
        with dib.Canvas do
          for y := 0 to Height - 1 do
            for x := 0 to Width - 1 do
              Pixels[x, y] := ShadePixel (Pixels[x, y], N + 1, SHADE_COUNT);

        pci := TPictureCollectionItem.Create (DXImageList.Items);
        pci.Name := tsn[i] + '_' + Chr (Ord ('0') + N);
        pci.PatternWidth := W;
        pci.PatternHeight := H;
        pci.TransparentColor := TRANSPARENT;
        pci.SystemMemory := TRUE;
        pci.Picture.Graphic := dib;
      end;
    end;


    SetLength (tsn, 0);
    
  end;
end;

procedure TGameWindow.NewGameState (gs: TGameState);
begin
  case CurGameState of
    gsInit: CurLevel := 1;
  end;
  CurGameState := gs;
  BackgroundColor := 0;
  case CurGameState of
    gsLevel:
      begin
        SetLength (Drops, 0);

        case CurLevel of
          1:
            begin
              SkyLayer := Sky1Map2Layer;
          //  BackLayer :=
              FrontLayer := Blocks1Level1Layer;
              BackgroundColor := $203040;
            end;
        end;
        XView := 0;
        YView := 0;

        SetLength (Candles, 0);
        InitMap;

        FPWind := 0;
        LightPerc := 100;
      end;
  end;
end;

procedure TGameWindow.RunFrame;
begin
  case CurGameState of
    gsInit:
      begin

        NewGameState (gsLevel);

      end;
    gsLevel:
      begin

        RunPlayer;
        MoveScreen;
        RunDrops;

      end;
  end;
end;

procedure TGameWindow.DrawFrame;
begin
  case CurGameState of
    gsInit:
      begin
        //
      end;
    gsLevel:
      begin

        DrawLayer (@SkyLayer, 0, 0, SCREEN_W, SCREEN_H, XView div 4, YView div 4, 0);

      // DrawLayer (@BackLayer, 0, 0, SCREEN_W, SCREEN_H, XView div 2, YView div 2, 0);

        DrawLayer (@FrontLayer, 0, 0, SCREEN_W, SCREEN_H, XView, YView, 3);

        DrawPlayer;

        DrawDrops;
      end;
  end;
end;

function TGameWindow.ReadMap (X, Y: Integer; Side: Boolean): Integer;
  var
    L: TLayerPtr;
    i: Integer;
begin
  L := @FrontLayer;

  if Side then
  begin
    if (X < 0) or (Y < 0) or (X >= L^.MapW) or (Y >= L^.MapH) then
    begin
      Result := $FF;
      Exit;
    end;
  end;

  X := (X + L^.MapW) mod L^.MapW;
  Y := (Y + L^.MapH) mod L^.MapH;

  i := IntArray (L^.BoundMapData)[Y * L^.MapW + X];

  Result := i;
end;

procedure TGameWindow.InitMap;
  var
    x, y, i: Integer;
    c: Integer;
begin
  for y := 0 to FrontLayer.MapH - 1 do
    for x := 0 to FrontLayer.MapW - 1 do
    begin
      c := ReadMap (x, y, FALSE);
      c := (c shr 8) and $FF;
      if c = $FF then  { map code $FF = player }
        InitPlayer (x, y);
      if c = $10 then  { map code $10 = candle }
      begin
        i := Length (Candles);
        SetLength (Candles, i + 1);
        with Candles[i] do
        begin
          XPos := x * FrontLayer.TileW + FrontLayer.TileW div 2;
          YPos := y * FrontLayer.TileH;
          Light := 0;
        end;
      end;
    end;
end;

procedure TGameWindow.DrawLayer (L: TLayerPtr; X, Y, W, H: Integer; XScroll, YScroll: Integer; Shade: Integer);
  var
    XPos, YPos, MapX, MapY, X1, Y1, X2, Y2: Integer;
    pci: TPictureCollectionItem;
    tile: Integer;
    Name: string;
begin
  Name := L^.TileSetID;
  if Pos (UpcaseStr (Name), SHADE_IMAGES) > 0 then
    if Shade < SHADE_COUNT - 1 then
      Name := Name + '_' + Chr (Ord ('0') + Shade);
  pci := DXImageList.Items.Find (Name);
  X1 := X + XScroll;
  Y1 := Y + YScroll;
  X2 := X + W;
  Y2 := Y + H;
  MapY := Y1 div L^.TileH;
  YPos := MapY * L^.TileH - YScroll;
  while YPos < Y2 do
  begin
    MapX := X1 div L^.TileW;
    XPos := MapX * L^.TileW - XScroll;
    while XPos < X2 do
    begin
      MapX := MapX mod L^.MapW;
      MapY := MapY mod L^.MapH;
      Tile := IntArray (L^.TileMapData)[MapY * L^.MapW + MapX];
      if Tile > 0 then
        pci.Draw (DXDraw.Surface, XPos, YPos, Tile - 1);
      Inc (MapX);
      Inc (XPos, L^.TileW);
    end;
    Inc (MapY);
    Inc (YPos, L^.TileH);
  end;
end;

procedure TGameWindow.FormActivate(Sender: TObject);
begin
  DXTimer.Enabled := TRUE;
end;

procedure TGameWindow.FormDeactivate(Sender: TObject);
begin
  DXTimer.Enabled := FALSE;
end;

procedure TGameWindow.FormKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #27 then
    Close;
end;

procedure TGameWindow.FormCreate(Sender: TObject);
begin
  ShowCursor (FALSE);

  CreateShades;

  NewGameState (gsInit);
end;

procedure TGameWindow.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key < 256 then
    Keys[Key] := TRUE;
  if (Key = VK_RETURN) and (ssAlt in Shift) then
    with DXDraw do
    begin
      Finalize;
      if doFullScreen in Options then
      begin
        RestoreWindow;
        BorderStyle := bsDialog;
        Options := Options - [doFullScreen];
      end
      else
      begin
        StoreWindow;
        BorderStyle := bsNone;
        Options := Options + [doFullScreen];
      end;
      Initialize;
    end;
end;

procedure TGameWindow.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key < 256 then
    Keys[Key] := FALSE;
end;

procedure TGameWindow.InitPlayer (MapX, MapY: Integer);
begin
  with Player do
  begin
    FPXPos := (MapX * FrontLayer.TileW + FrontLayer.TileW div 2) * FP;
    FPYPos := (MapY * FrontLayer.TileH + FrontLayer.TileH - 1) * FP;
    FPXVel := 0;
    FPYVel := 0;
    FPGravity := 32;
    Width := 64;
    Height := 128;
    AnimSpeed := 10;
    AnimPos := 0;

    XShift := 0;
    YShift := 1;
    BmpWidth := DXImageList.Items.Find ('MainChar').PatternWidth;
    BmpHeight := DXImageList.Items.Find ('MainChar').PatternHeight;

    State := psIdle;

    HDirection := -1;
    VDirection := -1;
    MoveScreen;
    HDirection := +1;
    VDirection := +1;
    MoveScreen;

  end;
end;

procedure TGameWindow.DrawPlayer;
  const
    MIRROR = 7;
  var
    Sprite: Integer;
    c, i, j, lx, ly, dx, dy: Integer;
    R: TRect;
begin
  for c := 0 to Length (Candles) - 1 do
    with Candles[c] do
    begin
      lx := XPos - XView;
      ly := YPos - YView;
      dx := Round ( (2) * Sin (FlareCounter / (2.2) ));
      dy := Round ( (2) * Sin (FlareCounter / (2.1) ));
      DXImageList.Items.Find ('Candle').Draw (DXDraw.Surface,
           lx - 8, ly - 2, 2 + Byte (Light < 100));
      DXImageList.Items.Find ('Flare').DrawAdd (DXDraw.Surface,
           MakeRect (lx - 32 - dy div 2, ly - 32 - dx div 2, 64 +dy, 64 +dx), 0,
             Light * (64 + dx+dy) div 100);
      Inc (FlareCounter);

      if Light = 100 then
      begin
        if CheckCol (XView + lx, YView + ly) then
          Light := 99;
      end;

      case Light mod 3 of
        0: Dec (Light, 3);
        2: Inc (Light, 3);
      end;
      if Light < 0 then
        Light := 0;
      if Light > 100 then
        Light := 100;
    end;

  with Player do
  begin
    i := udiv (FPXPos, FP) - XView;
    j := udiv (FPYPos, FP) - YView;

    R.Top := j - (BmpHeight - 1) + YShift;
    R.Left := i - BmpWidth div 2;
    R.Right := R.Left + BmpWidth;
    R.Bottom := R.Top + BmpHeight;

    Sprite := MIRROR * Byte (HDirection < 0);
    case State of
      psIdle: ;
      psWalk: Inc (Sprite, 1 + ((AnimPos div AnimSpeed) mod 4));
      psJump: Inc (Sprite, 5);
      psFall: Inc (Sprite, 6);
    end;

    lx := i + HDirection * (27);
    ly := j - (50);
    case State of
      psIdle: Dec (ly, 10);
      psWalk: Inc (ly, 5 * (AnimPos div AnimSpeed) mod 4);
      psJump: Dec (ly, 20);
      psFall: Inc (ly, 15);
    end;

    for c := 0 to Length (Candles) - 1 do
      with Candles[c] do
      begin
        if (Abs (XPos - (XView + lx)) <= MARGIN) and
           (Abs (YPos - (YView + ly)) <= MARGIN) then
        begin
          if LightPerc > Light then
            Light := 2
          else
            if Light > LightPerc then
              LightPerc := 2;
        end;
      end;

    if LightPerc = 100 then
    begin
      if CheckCol (XView + lx, YView + ly) then
        LightPerc := 99;
    end;

    if LightX = 0 then LightX := lx;
    if LightY = 0 then LightY := ly;
  //  if lx <> LightX then Inc (lx, (LightX - lx) div 2);
  //  if ly <> LightY then Inc (ly, (LightY - ly) div 2);
    if lx <> LightX then lx := (LightX + LightX + lx) div 3;
    if ly <> LightY then ly := (LightY + LightY + ly) div 3;

    dx := Round ( (3) * Sin (LightAnimPos / (3.7) ));
    dy := Round ( (3) * Sin (LightAnimPos / (4.2) ));

    DXImageList.Items.Find ('Flare').DrawAdd (DXDraw.Surface,
       MakeRect (lx - 120 -dx, ly - 120 -dy, 240 +2*dx, 240 +2*dy), 0,
                LightPerc * (128 + dx+dy) div 100);

    DXImageList.Items.Find ('Candle').Draw (DXDraw.Surface, lx - 8, ly - 8, Byte (LightPerc < 100));

    DXImageList.Items.Find ('Flare').DrawAdd (DXDraw.Surface,
       MakeRect (lx - 8 - dy div 2, ly - 8 - dx div 2, 16 +dy, 16 +dx), 0,
           LightPerc * (32 + dx+dy) div 100);

      //  DXImageList.Items.Find ('Flare').DrawAdd (DXDraw.Surface,
      //      MakeRect (50 + 22 + 5, 22 + 5, 256 - 10, 256 - 10), 0, 128);

    DXImageList.Items.Find ('MainChar').StretchDraw (DXDraw.Surface, R, Sprite);

  //  DXImageList.Items.Find ('MainChar').Draw (DXDraw.Surface, R.Left, R.Top, 0);

    //  DXImageList.Items.Find ('MainChar').Draw (DXDraw.Surface, 140, 140, CurAnim);

    dx := Round ( (2) * Sin (LightAnimPos / (2.3) ));
    dy := Round ( (2) * Sin (LightAnimPos / (2.2) ));

    DXImageList.Items.Find ('Flare').DrawAdd (DXDraw.Surface,
       MakeRect (lx - 64 +dx, ly - 64 +dy, 128 -2*dx, 128 -2*dy), 0,
       LightPerc * (64 -dx -dy) div 100);

        //  Rect (200 - 64, 150 - 64, 200 + 64 - 1, 150 + 64 - 1), 0, 64);

    LightX := lx;
    LightY := ly;

    Inc (AnimPos);
    Inc (LightAnimPos);

    case LightPerc mod 3 of
      0: Dec (LightPerc, 3);
      2: Inc (LightPerc, 3);
    end;
    if LightPerc < 0 then
      LightPerc := 0;
    if LightPerc > 100 then
      LightPerc := 100;
  end;
end;

procedure TGameWindow.RunPlayer;
  const
    FPACC = $40;
    FPDEC = $20;
    FPMAXXV = $200;
    FPMAXYV = $800;
    FPJUMP = $600;
begin
  with Player do
  begin
    if Keys[VK_SPACE] then
    begin
      if (not Spacebar) and (State in [psIdle, psWalk]) then
      begin
        FPYVel := -FPJUMP;
        State := psJump;
        Spacebar := TRUE;
      end;
    end
    else
      Spacebar := FALSE;

    if HDirection > 0 then
    begin
      if (Keys[VK_RIGHT]) or (FPXVel < 0) then
      begin
        if State in [psIdle] then
        begin  { start walking (right) }
          State := psWalk;
          AnimPos := 0;
        end;
        Inc (FPXVel, FPACC);
      end
      else
        if FPXVel > 0 then
        begin  { slow down }
          Dec (FPXVel, FPDEC);
          if FPXVel < 0 then
            FPXVel := 0;
        end;

      if FPXVel = 0 then
        if State = psWalk then
          State := psIdle;

      if FPXVel > FPMAXXV then
        FPXVel := FPMAXXV;

      if Keys[VK_LEFT] then
        HDirection := -1;
    end
    else
    begin
      if (Keys[VK_LEFT]) or (FPXVel > 0) then
      begin
        if State in [psIdle] then
        begin  { start walking (left) }
          State := psWalk;
          AnimPos := 0;
        end;
        Dec (FPXVel, FPACC);
      end
      else
        if FPXVel < 0 then
        begin  { slow down }
          Inc (FPXVel, FPDEC);
          if FPXVel > 0 then
            FPXVel := 0;
        end;

      if FPXVel = 0 then
        State := psIdle;

      if FPXVel < -FPMAXXV then
        FPXVel := -FPMAXXV;

      if Keys[VK_RIGHT] then
        HDirection := +1;
    end;

    if State = psFall then
      if FPYVel = 0 then
        State := psIdle;
    if State in [psJump, psIdle, psWalk] then
      if FPYVel > 3 * FPGravity then
        State := psFall;

    Inc (FPYVel, FPGravity);

    if FPYVel > FPMAXYV then
      FPYVel := FPMAXYV;
    if FPYVel < -FPMAXYV then
      FPYVel := -FPMAXYV;

    CheckLevelBounds;

    Inc (FPXPos, FPXVel);
    Inc (FPYPos, FPYVel);
  end;
end;

procedure TGameWindow.CheckLevelBounds;
  var
    W, H: Integer;
    xv, yv: Integer;
    CurX, NewX: Integer;
    CurX1, NewX1, CurX2, NewX2: Integer;
    CurY1, NewY1, CurY2, NewY2: Integer;
    CurMapX, NewMapX: Integer;
    CurMapX1, NewMapX1, CurMapX2, NewMapX2: Integer;
    CurMapY1, NewMapY1, CurMapY2, NewMapY2: Integer;
    Bnd, ChkBnd: Integer;
    x1, x2, y1, y2: Integer;
    i, j: Integer;
    hdir: Integer;

begin
  W := FrontLayer.TileW;
  H := FrontLayer.TileH;
  with Player do
  begin
    xv := FPXVel;
    yv := FPYVel;
    hdir := HDirection;
  {
    if xv > 0 then
      hdir := 1;
    if xv < 0 then
      hdir := -1;
  }
    CurX := udiv (FPXPos,      FP) + hdir * (Width div 2) - Byte (hdir = 1);
    NewX := udiv (FPXPos + xv, FP) + hdir * (Width div 2) - Byte (hdir = 1);
    CurMapX := udiv (CurX, W);
    NewMapX := udiv (NewX, W);

    CurX1 := udiv (FPXPos,      FP) - (Width div 2);
    NewX1 := udiv (FPXPos + xv, FP) - (Width div 2);
    CurX2 := CurX1 + Width - 1;
    NewX2 := NewX1 + Width - 1;

    CurMapX1 := udiv (CurX1, W);
    CurMapX2 := udiv (CurX2, W);
    NewMapX1 := udiv (NewX1, W);
    NewMapX2 := udiv (NewX2, W);

    CurY1 := udiv (FPYPos, FP) - (Height);
    CurY2 := udiv (FPYPos, FP);
    NewY1 := udiv (FPYPos + yv, FP) - (Height);
    NewY2 := udiv (FPYPos + yv, FP);

    CurMapY1 := udiv (CurY1, H);
    NewMapY1 := udiv (NewY1, H);
    CurMapY2 := udiv (CurY2, H);
    NewMapY2 := udiv (NewY2, H);

    if CurMapX <> NewMapX then
    begin
      y1 := NewMapY1;
      y2 := NewMapY2;
      if yv < 0 then
        y1 := CurMapY1;
      if yv > 0 then
        y2 := CurMapY2;

      if xv > 0 then
        ChkBnd := BND_LEFT
      else
        ChkBnd := BND_RIGHT;
      Bnd := 0;
      for j := y1 to y2 do
        Bnd := Bnd or ReadMap (NewMapX, j, TRUE);

      if Bnd and ChkBnd <> 0 then
        FPXVel := 0;
    end;

    if CurMapY1 <> NewMapY1 then
    begin
      x1 := NewMapX1;
      x2 := NewMapX2;
      if xv < 0 then
        x1 := CurMapX1;
      if xv > 0 then
        x2 := CurMapX2;
      if x1 > x2 then
        SwapInt (x1, x2);

      if yv >= 0 then
      begin
        j := NewMapY2;
        ChkBnd := BND_UPPER;
      end
      else
      begin
        j := NewMapY1;
        ChkBnd := BND_LOWER;
      end;

      Bnd := 0;

      for i := x1 to x2 do
        Bnd := Bnd or ReadMap (i, j, TRUE);

      if Bnd and ChkBnd <> 0 then
        FPYVel := 0;

    end;

  end;
end;

procedure TGameWindow.MoveScreen;
  var
    i, j: Integer;
begin
  with Player do
  begin
    i := udiv (FPXPos, FP) + HDirection * (Width div 2);
    j := udiv (FPYPos, FP) - Height * Byte (VDirection < 0);

   // if HDirection > 0 then
    begin
      if i + X_SCROLL_SIDE > XView + SCREEN_W then
        XView := i + X_SCROLL_SIDE - SCREEN_W;
    end;
   // if HDirection < 0 then
    begin
      if i - X_SCROLL_SIDE < XView then
        XView := i - X_SCROLL_SIDE;
    end;

   // if VDirection > 0 then
    begin
      if j + Y_SCROLL_SIDE > YView + SCREEN_H then
        YView := j + Y_SCROLL_SIDE - SCREEN_H;
    end;
   // if VDirection < 0 then
    begin
      if j - Y_SCROLL_SIDE < YView then
        YView := j - Y_SCROLL_SIDE;
    end;

    if XView < 0 then
      XView := 0;
    if YView < 0 then
      YView := 0;
    with FrontLayer do
    begin
    if XView + SCREEN_W > MapW * TileW then
      XView := MapW * TileW - SCREEN_W;
    if YView + SCREEN_H > MapH * TileH then
      YView := MapH * TileH - SCREEN_H;
    end;
  end;
end;

procedure TGameWindow.RunDrops;
  const
    FPMAXWIND = $100;
  var
    i, j: Integer;
begin
  for i := Length (Drops) - 1 downto 0 do
  begin
    with Drops[i] do
    begin
      Inc (FPXPos, FPXVel);
      Inc (FPYPos, FPYVel);
      FPXVel := (FPXVel + FPWind) div 2;

      if FPYPos > (YView + SCREEN_H + SCREEN_H div 10) * FP then
      begin
        for j := i + 1 to Length (Drops) - 1 do
          Drops[j - 1] := Drops[j];
        SetLength (Drops, Length (Drops) - 1);
      end;
    end;
  end;

  if Random (100) < 10 then
  begin
    i := Length (Drops);
    SetLength (Drops, i + 1);
    with Drops[i] do
    begin
      FPXPos := (XView + Random (3 * SCREEN_W) - SCREEN_W) * FP;
      FPYPos := (YView - SCREEN_H div 10) * FP;
      FPXVel := FPWind;
      FPYVel := $100 + Random ($100);
      Shade := 1; // Random (3);
    end;
  end;

  FPWind := FPWind + Random (8) - Random (8);
  if FPWind > FPMAXWIND then
    FPWind := FPMAXWIND;
  if FPWind < -FPMAXWIND then
    FPWind := -FPMAXWIND;
end;

procedure TGameWindow.DrawDrops;
  var
    i: Integer;
    x, y: Integer;
    pci: TPictureCollectionItem;
    w, h: Integer;
begin
 // pci := DXImageList.Items.Find ('Rain');
  pci := DXImageList.Items.Find ('Rain_' + Chr (Ord ('0') + Random (SHADE_COUNT)));
  w := pci.PatternWidth;
  h := pci.PatternHeight;
  for i := 0 to Length (Drops) - 1 do
  begin
    with Drops[i] do
    begin
      x := udiv (FPXPos, FP) - w div 2;
      y := udiv (FPYPos, FP) - h div 2;

      pci.Draw (DXDraw.Surface, x - XView, y - YView, Shade * 3 + Random (3));

    end;
  end;
end;

procedure TGameWindow.FormDestroy(Sender: TObject);
begin
  SetLength (Drops, 0);
  SetLength (Candles, 0);
end;

function TGameWindow.CheckCol (x, y: Integer): Boolean;
  var
    i, xp, yp: Integer;
begin
  Result := FALSE;
  for i := Length (Drops) - 1 downto 0 do
  begin
    with Drops[i] do
    begin
      xp := udiv (FPXPos, FP);
      yp := udiv (FPYPos, FP);
      if (Abs (xp - x) <= MARGIN) and (Abs (yp - y) <= MARGIN) then
        Result := TRUE;
    end;
  end;
end;

end.

